/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#include "Atomic.h"
#include <float.h>
#include "Coupled.h"
#include "ODEVS_Exception.h"
#include "ClassAdmin.h"

using namespace ODEVS;

IMPLEMENT_ABSTRACT_CLASS1(Atomic, "Atomic", 20060806, Model);

string Atomic::Get_q(bool remaining) const
{
	string str = "("+Name+":"+Get_s()+",";
	char tmp[100];
	TimeSpan t_s = TimeNext()-TimeLast();
	if(TimeNext() == DBL_MAX) // passive
		if(remaining)
			sprintf(tmp, "inf,inf)");
		else // elapsed time
			sprintf(tmp, "%.3lf,inf)", TimeElapsed());
	else // active 
		if(remaining)
			sprintf(tmp, "%.3lf,%.3lf)", t_s, TimeRemaining());
		else
			sprintf(tmp, "%.3lf,%.3lf)", TimeElapsed(), t_s);
	str += tmp;
	return str;
}

void Atomic::when_receive_init()  
{ 
	init();
	t_Last = TimeCurrent();
	t_Next = TimeCurrent()+ta();
}


void Atomic::when_receive_star(bool print_event)
{
	if(IsEqual(t_Next, t_Current, 1E-5)) {
		PortValue y=lambda();
		if(print_event)
			cout << "!" << GetHierName(NULL,this) << "." << y.ToString();

		delta_int();
		if(Parent)
			Parent->when_receive_y(*this, y,print_event);

		t_Last = TimeCurrent();
		if(ta() == DBL_MAX)
			t_Next = DBL_MAX;
		else
			t_Next = t_Current+ta();
	}else
		THROW_ODEVS_EXCEPTION("Schedule Violation!");
}


void Atomic::when_receive_phi()
{
	if(t_Last <= t_Current && t_Current <= t_Next) {
		Phi();
	}
	else
		THROW_ODEVS_EXCEPTION("Schedule Violation!");
}

void Atomic::when_receive_x(const PortValue& x, bool print_event)
{
	if(t_Last <= t_Current && t_Current <= t_Next) {
		rho = false; 
		if(print_event)
			cout << ",?" << GetHierName(NULL,this) << "." << x.ToString();

		delta_ext(x); 
		if(rho == true) {// reschedule is needed.
			t_Last = TimeCurrent();
			if(ta() == DBL_MAX)
				t_Next = DBL_MAX;
			else
				t_Next = t_Current+ta();
		}
	}else
		THROW_ODEVS_EXCEPTION("Schedule Violation!");
}

