/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#pragma once
 
#include "Model.h" 
#include "ODEVS_Export.h"

using namespace std;

namespace ODEVS {

/// base class of atomic DEVS models
class ODEVS_EXP Atomic: public Model
{
public:
	/*-- 5 characteristic functions  --*/
	virtual void init() = 0;
	virtual void delta_ext(const PortValue& x) = 0;
	virtual void delta_int() = 0;
	virtual PortValue lambda() const = 0;
	virtual TimeSpan ta() const = 0;
    //-----------------------------------------------
	//---- function for tracing state ----------------
	virtual string Get_s() const { return string(); }

	/* return string of (s,ta(s),t_r) if remaining==true
	   return string of (s,ta(s),t_e) otherwise */
	/*virtual*/ string Get_q(bool remainig=true) const ;

	//-- function for analytical continuous change
	//-- Phi: S_c * T -> S_c 
	//-- If the model has no continuous variable, we don't care of it.
	virtual void Phi() {}

	//-- virtual destructor 
	virtual ~Atomic() {}
	
	bool Get_rho() const { return rho; }
	//-- function making reschedule of an external transition
	void x_RescheduleMe(bool flag=true) { rho = flag; }

protected:


	Atomic(const std::string& name): 
		Model(name), rho(false) {}

	/*-- functions called by simulation algorithm --*/ 
	/*virtual*/ void when_receive_init(); 
	/*virtual*/ void when_receive_star(bool print_event); 
	/*virtual*/ void when_receive_phi(); 
	/*virtual*/ void when_receive_x(const PortValue& x,bool print_event);   

	bool rho;
	DECLARE_CLASS();
};
  
}// end of namespace of ODEVS 
