/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#pragma once
#pragma warning(disable:4251)

#include "Model.h"
#include "PortValue.h"
#include "ODEVS_Export.h"
#include <deque>
#include <list>
#include <map>


namespace ODEVS {
  
//-- a pair struct of Model and Port
struct ModelPort
{
	Model* model;
	Port  port;

	ModelPort(Model* md=NULL, Port pt = ""):model(md), port(pt){}
	bool operator<(const ModelPort& ob) const
	{
		if(model < ob.model) return true;
		if(model == ob.model && port < ob.port) return true;
		return false;
	}
};
 
typedef multimap<ModelPort,ModelPort> CoupleMap; // ITC;

//-- class of DEVS Network 
class ODEVS_EXP Coupled: public Model
{
public:
	// constructor
	Coupled(const string& name=""): Model(name) {}
	// destructor
	virtual ~Coupled();

	// modelling related 
	void AddModel(Model* md); 
	Model* GetModel(const string& name) const;
	void Couple(Model* smd,  InputPort spt, Model* dmd, InputPort dpt); // EIC
	void Couple(Model* smd, OutputPort spt, Model* dmd, InputPort dpt); // IT
	void Couple(Model* smd, OutputPort spt, Model* dmd, OutputPort dpt);// EOC

	/* methods for observing status of model */
	/*virtual*/ string Get_q(bool remaining=true) const ;
	//-- print all couplings in a hierarchical manner
	void PrintCouplings() const;

	//---- 
	CoupleMap& Couples() { return C; }
	const CoupleMap& Couples() const { return C; }
	deque<Model*>& Children() { return M; }
	const deque<Model*>& Children() const { return M; }

	void GetInfluencees_x(InputPort x, deque<pair<Atomic*, InputPort> >& M_X) const;
	void GetInfluencees_y(Model* child, OutputPort y, deque<pair<Atomic*, InputPort> >& M_X) const;

protected:
	deque<Model*>     M;
	deque<Model*> Schedule;
	
	void Update_tL_tN() ;

	CoupleMap  C;

	//-- simulation run related 
	/// function called when simulation initialized
	/*virtual*/ void when_receive_init();
	/// function called when schedule-time reached
	/*virtual*/ void when_receive_star(bool print_event);
	/// function called when continuous time reached
	/*virtual*/ void when_receive_phi(); // print_event for debugging
	/// function called when an external input "x" comes
	/*virtual*/ void when_receive_x(const PortValue& x,bool print_event);
	void when_receive_y(Model& child, const PortValue& y,bool print_event);
friend class Atomic;

	DECLARE_CLASS();

};
 
}// end of namespace of ODEVS 
