/************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>      * 
 *                                                                      *
 * This library is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU Lesser General Public           *
 * License as published by the Free Software Foundation; either         *
 * version 2.1 of the License, or (at your option) any later version.   *
 *                                                                      *
 * This library is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     *
 * Lesser General Public License for more details.                      *
 *                                                                      *
 * You should have received a copy of the GNU Lesser General Public     *
 * License along with this library; if not, write to the Free Software  *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,*  
 * USA.                                                                 * 
*************************************************************************/
#pragma once
#include <Atomic.h>
#include <deque>
#include <vector>
#include "Client.h"

using namespace ODEVS;

const string IDLE = "Idle";
vector<string> SENDTO; 

class Buffer: public Atomic
{
public:
	vector<OutputPort> out;  
	InputPort in;
	vector<InputPort> pull;
	
protected:
	std::deque<Client*> m_Clients;
	string m_phase;
	vector<bool> m_OAvail;
	int			 m_Osize;
public:
	Buffer(const string& name="", int osize=1): Atomic(name), m_Osize(osize), in("in")
	{
		for(int i=0; i<m_Osize; i++){
			char tmp[10]; 
			::sprintf_s(tmp, "out%d", i); out.push_back(OutputPort(tmp));
			::sprintf_s(tmp, "pull%d", i); pull.push_back(InputPort(tmp));
			::sprintf_s(tmp, "SendTo%d", i); SENDTO.push_back(string(tmp));
		}

		init();
	}
	virtual ~Buffer() { init(); }

    /*virtual*/ void init()
	{
		m_phase = IDLE;
		m_OAvail.clear(); // clear first
		for(int i=0; i<m_Osize; i++)
			m_OAvail.push_back(true); // add variable

		while(m_Clients.empty() == false)
		{
			Client* cl = m_Clients[0]; 
			m_Clients.pop_front();
			delete cl;
		}
	}

	/*virtual*/ void delta_ext(const PortValue& x) 
	{
		if(m_phase == IDLE){
			if(x.port == in){ //receiving a client
				Client* client = dynamic_cast<Client*>(x.value);
				m_Clients.push_back(client);
				for(int i=0; i < m_Osize; i++){// select server
					if(m_OAvail[i] == true){// server i is available
						m_phase = SENDTO[i]; // we're going to send a client server_i
						m_OAvail[i]=false;// mark server_i non-available fron now on.
						x_RescheduleMe(); // reschedule as active
						break;
					}
				}
			}else // receiving a pull signal 
			{
				for(int i=0; i < m_Osize; i++){  
					if(x.port == pull[i]){ // server_i is available
						m_phase = SENDTO[i];// we're going to send a client server_i
						m_OAvail[i]= false;// mark server_i non-available fron now on.
						x_RescheduleMe(); // reschedule as active
						break;
					}
				}
			}
		}else // when sending somewhere 
		{
			 if(x.port == in){ // receiving a client
				Client* client = dynamic_cast<Client*>(x.value);
				assert(client);
				m_Clients.push_back(client);
			}else // receiving a pull signal  
			{
				for(int i=0; i < m_Osize; i++){  
					if(x.port == pull[i]){// server_i is available
						m_OAvail[i]= true; // server_i gets available
						break;
					}
				}
			}
			// notice that we DON'T call x_RescheduleMe()! => continue schedule
		}
	}
	/*virtual*/ void delta_int() 
	{
		m_Clients.pop_front(); // changing state in delta_int() not in lambda()
		if(m_Clients.empty())
			m_phase = IDLE;
		else 
		{	
			for(int i=0; i < m_Osize; i++) {// finding an available server 
				if(m_OAvail[i] == true) { // server_i is available
					m_phase = SENDTO[i];// we're going to send a client server_i
					m_OAvail[i]= false;// we're using server_i from now on.
					return;
				}
			}
			m_phase = IDLE;
		}
	}

	/*virtual*/ PortValue lambda() const
	{ 
		for(int i=0; i < m_Osize; i++){
			 if(m_phase == SENDTO[i]) 
				 return PortValue(out[i], m_Clients[0]); // return dollar
		}
		return PortValue(); // slient event
	}

	/*virtual*/ Time ta() const 
	{ 
		if(m_phase == IDLE)
			return DBL_MAX;
		else
			return 2.0;
	}

	/*virtual*/ string Get_s() const 
	{
		char buffer[10]; _itoa_s((int)m_Clients.size(), buffer, 9);
		string s = m_phase+",#J:"+buffer+",AVs:";

		for(int i=0; i<m_Osize; i++)
		{
			if(i>0) s+=",";
			if(m_OAvail[i] == true)
				s += "1";
			else
				s += "0";
		}
		return s; 
	}
};
