/************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>      * 
 *                                                                      *
 * This library is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU Lesser General Public           *
 * License as published by the Free Software Foundation; either         *
 * version 2.1 of the License, or (at your option) any later version.   *
 *                                                                      *
 * This library is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     *
 * Lesser General Public License for more details.                      *
 *                                                                      *
 * You should have received a copy of the GNU Lesser General Public     *
 * License along with this library; if not, write to the Free Software  *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,*  
 * USA.                                                                 * 
*************************************************************************/
#pragma once
#include <Atomic.h>
#include <deque>
#include <vector>
#include "Client.h"

using namespace ODEVS;

//const string IDLE = "Idle";
const string BUSY = "Busy";
const string COLLIDED = "Collided";

class Server: public Atomic
{
public:
	OutputPort out;  
	InputPort in;
	
protected:
	string m_phase;
	Client*   m_pClient;
public:
	Server(const string& name=""): Atomic(name), 
		in("in"), out("out"), m_pClient(NULL)
	{
		init();
	}
	virtual ~Server() { init(); }

    /*virtual*/ void init()
	{
		m_phase = IDLE;
		if(m_pClient)
			delete m_pClient;
		m_pClient=NULL;
	}

	/*virtual*/ void delta_ext(const PortValue& x) 
	{
		if(x.port == in){ //receiving a client
			if(m_phase == IDLE){
				m_pClient = dynamic_cast<Client*>(x.value);
				assert(m_pClient);
				m_phase = BUSY;
				x_RescheduleMe();
			}else if( m_phase == BUSY)
			{
				m_phase = BUSY;
				x_RescheduleMe();
			} // already collided,
		} 
	}
	/*virtual*/ void delta_int() 
	{
		m_pClient = NULL; // don't delete that. sink will delete that.
		m_phase = IDLE;
	}

	/*virtual*/ PortValue lambda() const
	{ 
		return PortValue(out, m_pClient); 
	}

	/*virtual*/ Time ta() const 
	{ 
		if(m_phase == IDLE)
			return DBL_MAX;
		else if(m_phase == BUSY) {
			assert(m_pClient);
			return m_pClient->TimeProcess;
		}
		else
			return DBL_MAX;
	}

	/*virtual*/ string Get_s() const 
	{
		return m_phase; 
	}
};
