/************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>      * 
 *                                                                      *
 * This library is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU Lesser General Public           *
 * License as published by the Free Software Foundation; either         *
 * version 2.1 of the License, or (at your option) any later version.   *
 *                                                                      *
 * This library is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     *
 * Lesser General Public License for more details.                      *
 *                                                                      *
 * You should have received a copy of the GNU Lesser General Public     *
 * License along with this library; if not, write to the Free Software  *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,*  
 * USA.                                                                 * 
*************************************************************************/
#include <Atomic.h>
#include <Coupled.h>

#include <SRTEngine.h>
#include <iostream>
#include <math.h>

using namespace std;
using namespace ODEVS;

const OutputPort OP1="send1";
const OutputPort OP2="send2";
const OutputPort OP="send";

const InputPort IP1="receive1";
const InputPort IP2="receive2";
const InputPort IP="receive";

const string WAIT = "wait";
const string SEND = "send";

class Player: public Atomic
{
protected:
	string    m_phase;      //
	bool      m_width_ball; 
	double    at_time; // attack_time
public:
    Player(const string& name="", double ttime=1.0, bool with_ball=false): Atomic(name), 
		m_phase(WAIT), m_width_ball(with_ball), at_time(ttime) {}
	/*virtual*/ void init()
		{ 
			if(m_width_ball)
				m_phase = SEND;
			else
				m_phase = WAIT;
		}
	/*virtual*/ void delta_ext(const PortValue& x)
	{
		if(x.port == IP)
		{
			if(m_phase == WAIT) {
				m_phase = SEND;
				x_RescheduleMe();
			}else {  
				cout << "Do we have more than one ball?";
				exit(0);
			}
		}
	}
  	/*virtual*/ void delta_int()  
	{
		if(m_phase == SEND) {
		    m_phase = WAIT;
		}
	}

	/*virtual*/ PortValue lambda() const
	{
		return PortValue(OP, NULL);
	}
	/*virtual*/ TimeSpan ta() const 
	{
		if(m_phase == SEND)
			return at_time;
		else
			return DBL_MAX;
	}
	/*virtual*/ string Get_s() const
	{
		return m_phase;
	}
}; 

Coupled* MakeTeam(const string& name, double attime, bool active)
{
	//-- One team consists of player 1 and player 2 ------
	Coupled* Team =  new Coupled(name);
	Player* p1 = new Player("p1", attime, active); 	
	Player* p2 = new Player("p2", attime, false); 	
	Team->AddModel(p1);
	Team->AddModel(p2);
	//-- External Output Coupling  
	Team->Couple(p1, OP, Team, OP1);
	Team->Couple(p2, OP, Team, OP2);

	//-- External Input Coupling 
	Team->Couple(Team, IP1, p1, IP);
	Team->Couple(Team, IP2, p2, IP);
	return Team;
}

Coupled* MakeDoubleMatch(const string& name, double attime)
{
	//-- DoubleMatch consists of two teams which have two player, respectively 
	Coupled* DoubMatch =  new Coupled(name);
	Coupled* TeamA = MakeTeam("Team_A", attime, true);
	Coupled* TeamB = MakeTeam("Team_B", attime, false);

	DoubMatch->AddModel(TeamA);
	DoubMatch->AddModel(TeamB);
	//-- Internal Coupling  
	//-- the receiver returns the ball oponent player who didn't give the ball
	//-- to the receiver. So A.p1->B.p2->A.p2->B.p1->A.p1 ... 
	DoubMatch->Couple(TeamA, OP1, TeamB, IP2);
	DoubMatch->Couple(TeamB, OP2, TeamA, IP2);
	DoubMatch->Couple(TeamA, OP2, TeamB, IP1);
	DoubMatch->Couple(TeamB, OP1, TeamA, IP1);
	
	return DoubMatch;
}

Coupled* MakeGym(const string& name)
{
	Coupled* Gym = new Coupled(name);
	Gym->AddModel(MakeDoubleMatch("Match1", 0.5)) ;  
	Gym->AddModel(MakeDoubleMatch("Match2", 0.7)) ;  
	return Gym;
}

void main(void)
{
	Coupled* Gym = MakeGym("Gym");
	SRTEngine simEngine(*Gym);  
	simEngine.RunConsoleMenu();
	delete Gym;
}

