/************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>      * 
 *                                                                      *
 * This library is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU Lesser General Public           *
 * License as published by the Free Software Foundation; either         *
 * version 2.1 of the License, or (at your option) any later version.   *
 *                                                                      *
 * This library is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     *
 * Lesser General Public License for more details.                      *
 *                                                                      *
 * You should have received a copy of the GNU Lesser General Public     *
 * License along with this library; if not, write to the Free Software  *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,*  
 * USA.                                                                 * 
*************************************************************************/
#include <Atomic.h>
#include <Coupled.h>
#include <SRTEngine.h>
#include <iostream>
#include <math.h>

using namespace std;
using namespace ODEVS;


const OutputPort OP="send";
const InputPort IP="receive";

const string WAIT = "wait";
const string SEND = "send";

class Player: public Atomic
{
protected:
	string    m_phase;      //
	bool      m_width_ball; 
public:
    Player(const string& name="", bool with_ball=false): Atomic(name), 
		m_phase(WAIT), m_width_ball(with_ball) {}
	/*virtual*/ void init()
	{ 
		if(m_width_ball)
			m_phase = SEND;
		else
			m_phase = WAIT;
	}
    /*virtual*/ void delta_ext(const PortValue& x)
	{
		if(x.port == IP)
		{
			if(m_phase == WAIT) {
				m_phase = SEND;
				x_RescheduleMe();
			}else {  
				cout << "Do we have more than one ball?";
				exit(0);
			}
		}
	}
  	/*virtual*/ void delta_int() // --------- (3) 
	{
		if(m_phase == SEND) {
		    m_phase = WAIT;
		}
	}
	/*virtual*/ PortValue lambda() const
	{
		return PortValue(OP, NULL);
	}
	/*virtual*/ TimeSpan ta() const 
	{
		if(m_phase == SEND)
			return 0.2;
		else
			return DBL_MAX;
	}
	/*virtual*/ string Get_s() const
	{
		return m_phase;
	}
}; 

const string IDLE="IDLE";
const string AtoB="PING";
const string BtoA="PONG";


class PPTable: public Atomic
{
protected:
	string    m_phase;      //

public:
	InputPort  ipA, ipB;
	OutputPort opA, opB;

    PPTable(const string& name=""): Atomic(name), m_phase(IDLE),
		ipA("ipA"), ipB("ipB"),opA("opA"), opB("opB") {}
	/*virtual*/ void init()
	{ 
		m_phase = IDLE;
	}
    /*virtual*/ void delta_ext(const PortValue& x)
	{
		if(m_phase == IDLE) {
			if(x.port == ipA)
			{
				m_phase = AtoB;
				x_RescheduleMe();
			}else if(x.port == ipB) {
				m_phase = BtoA;
				x_RescheduleMe();
			}else{
				cout << "Do we have more than one ball?";
				exit(0);
			}
		}
		x_RescheduleMe(); // Notice Here !
	}
  	/*virtual*/ void delta_int() // --------- (3) 
	{
		if(m_phase == AtoB || m_phase == BtoA) {
		    m_phase = IDLE;
		}
	}
	/*virtual*/ PortValue lambda() const
	{
		if(m_phase == AtoB)
			return PortValue(opB, NULL);
		else if(m_phase == BtoA)
			return PortValue(opA, NULL);
		return PortValue();
	}
	/*virtual*/ TimeSpan ta() const 
	{
		if(m_phase == AtoB || m_phase == BtoA)
			return 1.0;
		else
			return DBL_MAX;
	}
	/*virtual*/ string Get_s() const
	{
		return m_phase;
	}
}; 

Coupled* MakePingPongGame(const string& name)
{
	//-- PingPong Game: Coupled Model ------ (1) 
	Coupled* PingPong =  new Coupled(name);
	Player* A = new Player("A", true); 	
	Player* B = new Player("B", false); 
	PPTable* TB = new PPTable("Table"); 
	PingPong->AddModel(A);
	PingPong->AddModel(TB);
	PingPong->AddModel(B);
	//-- Internal Coupling -------- (2) 
	PingPong->Couple(A, OP, TB, TB->ipA);
	PingPong->Couple(TB, TB->opB, B, IP);

	PingPong->Couple(B, OP, TB, TB->ipB);
	PingPong->Couple(TB, TB->opA, A, IP);
	return PingPong;
}


void main(void)
{
	Coupled* PingPong = MakePingPongGame("Match") ; //-- simulation model
	SRTEngine simEngine(*PingPong); // plug-in sim. model to sim. engine 
	simEngine.RunConsoleMenu();
	delete PingPong;
}

