/************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>      * 
 *                                                                      *
 * This library is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU Lesser General Public           *
 * License as published by the Free Software Foundation; either         *
 * version 2.1 of the License, or (at your option) any later version.   *
 *                                                                      *
 * This library is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     *
 * Lesser General Public License for more details.                      *
 *                                                                      *
 * You should have received a copy of the GNU Lesser General Public     *
 * License along with this library; if not, write to the Free Software  *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,*  
 * USA.                                                                 * 
*************************************************************************/
#include <Atomic.h>
#include <SRTEngine.h>
#include <iostream>
#include <math.h>
#include <assert.h>

using namespace std;
using namespace ODEVS;

const string IDLE="Idle";
const string WAIT="Wait";
const string O_PEPSI="O_pepsi";
const string O_COCA="O_coca";

class VM: public Atomic
{
public:
	const InputPort  id, // input for dollar
					 ip, // input for pepsi botton
					 ic; // input for coca botton
	const OutputPort od, // output for dollar 
					 op, // output for a can of pepsi
					 oc; // output for a can of coca 

	VM(const string& name=""): Atomic(name), 
		id("dollar"), ip("pepsi_btn"), ic("coca_btn"),
		od("dollar"), op("pepsi"), oc("coca") 
	{ 
		init();
	}

    /*virtual*/ void init()
	{
		m_phase = IDLE;
	}

	/*virtual*/ void delta_ext(const PortValue& x) 
	{
		if(m_phase == IDLE && x.port == id){
			m_phase = WAIT;
			x_RescheduleMe();
		} else if(m_phase == WAIT && x.port == ip) {
			m_phase = O_PEPSI;
			x_RescheduleMe();
		} else if(m_phase == WAIT && x.port == ic) {
			m_phase = O_COCA;
			x_RescheduleMe();
		}
	}

	/*virtual*/ void delta_int() 
	{
		if(m_phase == WAIT || m_phase == O_PEPSI || m_phase == O_COCA)
			m_phase = IDLE;
		else
			assert(0);
	}
	/*virtual*/ PortValue lambda() const
	{ 
		if(m_phase == WAIT)
			return PortValue(od); // return dollar
		else if(m_phase == O_PEPSI)
			return PortValue(op);
		else //if(m_phase == O_COCA)
			return PortValue(oc);
	}
	/*virtual*/ Time ta() const 
	{ 
		if(m_phase == WAIT)
			return 15;
		else if(m_phase == O_PEPSI)
			return 2;
		else if(m_phase == O_COCA)
			return 2;
		else
			return DBL_MAX;
	}
	/*virtual*/ string Get_s() const 
	{
		return m_phase; 
	}
protected:
	string m_phase;
};


//-- call back function to inject an event to simulator
PortValue InjectMsg()
{
	string input;
	cout << "[d]ollar [p]epsi_botton [c]oca_botton > " ;
	cin >> input;
	if(input == "d")
		return PortValue("dollar");
	else if(input == "p")
		return PortValue("pepsi_btn");
	else if(input == "c")
		return PortValue("coca_btn");
	else {
		cout <<"Invalid input! Try again! \n";
		return PortValue();
	}
}

void main( void )
{
	VM* vm = new VM("VM") ; //-- simulation model
	SRTEngine simEngine(*vm, InjectMsg); // plug-in sim. model to sim. engine 
	simEngine.RunConsoleMenu();
	delete vm;
}

