/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#pragma once

#include <string>
#include "ODEVS_Export.h"

namespace ODEVS {

typedef std::string Port;

class InputPort: public Port
{
public:
	InputPort(){}
	InputPort(const std::string& name): Port(name){}
};

class OutputPort: public Port
{
public:
	OutputPort(){}
	OutputPort(const std::string& name): Port(name){}
};

/*-- abstract class of Value from which all the message value
     should be derived --*/
class Value
{
protected:
	Value(){}
public:
	//-- copy me to another instance
	virtual Value* Clone() const {return NULL;}
	//-- convert to string
	virtual std::string ToString() const {return std::string(); }
};

class PortValue
{
public:
	//-- 
	Port  port;
	//-- to use, cast type!
	Value* value;

	//------------------------- constructors -------------------------
	PortValue(Port p="", Value* v=NULL){ SetPortValue(p, v); }
	PortValue(const PortValue& ob){ SetPortValue(ob.port, ob.value); }
	//----------------------------------------------------------------
	//---------------------- set or assign operator ------------------
	void SetPortValue(Port p, Value* v=NULL) { port = p; value = v;}
	const PortValue& operator=(const PortValue& ob)
			{ SetPortValue(ob.port, ob.value); return *this; }
	std::string ToString() const 
	{
		std::string str = port;
		if(value)
			str += ":"+value->ToString();
		return str;
	}
	//----------------------------------------------------------------
};

} // end of namespace of ODEVS
