/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#include "SRTEngine.h"
#include "ODEVS_Exception.h"
#include "ClassAdmin.h"
#include <iostream>
#include <math.h>
#include "Coupled.h"

#ifdef SLEEP 
#undef SLEEP
#endif

#if defined(WIN32)
#include <windows.h> // <afx.h> // <windows.h>  
#define CLEAR_SCREEN() system("cls")
#define SLEEP(n) Sleep(n)
#else
#include <unistd.h>
#define SLEEP(n) sleep(n)
#endif

using namespace ODEVS;

#if defined(WIN32)
	HANDLE    m_hThread(0L);
//friend UINT RootCoordiThread(LPVOID lpParam);
#else
	pthread_t m_hThread(0L);
//friend void* RootCoordiThread(void* lpParam);
#endif

#define MILLISECS_PER_DAY 24*60*60*1000

DWORD day_start = 0;

unsigned long milliseconds()
{
    SYSTEMTIME t_now;
	GetSystemTime(&t_now);
	return  t_now.wMilliseconds
			+ 1000*(60*(60*t_now.wHour + t_now.wMinute)+t_now.wSecond)+
			+ (t_now.wDay-day_start)*MILLISECS_PER_DAY;
}


/**
The ThreadProc function acts as the starting point for executing activity
threads.
*/
#if defined(WIN32)
DWORD WINAPI ThreadProc(LPVOID activity)
#else
void* ThreadProc(void* lpParam)
#endif
{
	SRTEngine* a = (SRTEngine*)activity;
	a->Loop();
	return 0;
}


SRTEngine::SRTEngine(Model& model, CallBack cbf, Time ending_t): 
m_model(model), call_back_method(cbf), m_ending_time(ending_t)//, m_bigmouth(bigmouth)  
{
	m_StepMode = GO_THROUGH;  
	m_time_scale = 1; //v default: real time
    Init();
	
	SYSTEMTIME t_now;
	GetSystemTime(&t_now);
	day_start = t_now.wDay;
	print_state_change = SRTEngine::P_Q_t_elapsed;  
	b_acs = true; // -- animation one 
	m_animation_interval = 0.25;

 //   simEngine.SetTimeScale(DBL_MAX); // virtual time simulation (fastest)
	//simEngine.Set_dtmode(SRTEngine::P_NONE); //SRTEngine::P_Q_t_elapsed);// print execution run of elapsed time
	//simEngine.SetAnimationFlag(false); 

#if defined(WIN32)
	DWORD dwThreadId;
	m_hThread = ::CreateThread(NULL, 0, 
					(LPTHREAD_START_ROUTINE)ThreadProc, 
					(void*)this, 0, &dwThreadId);
#else
	if(::pthread_create(&m_hThread, NULL,
			ThreadProc, (void*)this) != 0) 
	m_hThread = 0;// it might by necessary ... but...
#endif
}  


void SRTEngine::StampClockTime()
{
	m_PreviousClockTime = milliseconds(); // remember the starting clock time
}
 
void SRTEngine::SetTimeScale(double ts)
{
    if (ts == 0.0 || ts < 0.0)
		std::cout << ts << " should be grater than zero!\n";
    else
        m_time_scale = ts;
}

void SRTEngine::ScheduleExecution()
{
	if (m_model.TimeCurrent() >= m_ending_time){
		Model::t_Current = m_ending_time;
		cout << "The current time has reached to the simulation ending time." << endl;
		cout << m_menu;
		m_bPauseReq = true; 
	}
	else if (m_model.TimeNext() == DBL_MAX)
	{
		cout << "All models are passive." << endl;
		cout << m_menu;
		m_bPauseReq = true; 
	}
	else {
		try{
			if(b_acs)
				CLEAR_SCREEN();
			else
				cout << endl;
			if(print_state_change != P_NONE)
				cout << m_model.Get_q((print_state_change==P_Q_t_remaining)) 
					 << endl << " --({" ;
			m_model.when_receive_star((print_state_change != P_NONE)); 
			if(print_state_change != P_NONE){
				cout << "}," << m_model.TimeCurrent()<<")--> " << endl
					 << m_model.Get_q((print_state_change==P_Q_t_remaining)) << endl;
			}
		}
		catch(Exception& e)
		{
			m_bExitReq = true;
			cout << e << endl;
			throw e;
		}
		if ( m_StepMode == STEP_BY_ONE_SCHEDULE) {// stop looping
			m_bPauseReq = true;
			cout << m_menu;
		}
	}
}
//------------------- (9) -----------------------
void SRTEngine::AnimationExecution()
{
	if (m_model.TimeCurrent() >= m_ending_time){
		Model::t_Current = m_ending_time; // adjust to the ending time
		m_model.when_receive_phi(); 
		CLEAR_SCREEN();
		cout << m_model.Get_q((print_state_change==P_Q_t_remaining)) << " at " << m_model.TimeCurrent() << endl;
		cout << "The current time has reached to the simulation ending time." << endl;
		cout << m_menu;
		m_bPauseReq = true;
	}
	else {
		try{
			m_model.when_receive_phi(); 
			CLEAR_SCREEN();
			cout << m_model.Get_q((print_state_change==P_Q_t_remaining)) << " at " << m_model.TimeCurrent() << endl
				 << m_menu;
		}
		catch(Exception& e)
		{
			m_bExitReq = true;
			cout << e << endl;
			throw e;
		}
	}
}
//++++++++++++++++++++++++++++ (6) ++++++++++++++++++++++++++++++
void SRTEngine::TransmitExtEvent()
{
	/*if (m_model.TimeCurrent() >= m_ending_time){
		cout << "The current time has reached to the simulation ending time." << endl;
		m_bPauseReq = true; 
	}
	else*/ 
	{
		//Model::t_Current = m_model.TimeNext(); // adjust to the schedule

		while(m_ExtX.empty() == false)
		{
			PortValue pv = m_ExtX.front(); m_ExtX.pop_front();
			try{
				if(b_acs)
					CLEAR_SCREEN();
				else
					cout << endl;
				if(print_state_change != P_NONE)
					cout << m_model.Get_q((print_state_change==P_Q_t_remaining)) 
					<< endl << " --({?" << pv.port ;
				m_model.when_receive_x(pv,(print_state_change != P_NONE));
				if(print_state_change != P_NONE){
					cout << "}," << m_model.TimeCurrent()<<")--> " << endl
						 << m_model.Get_q((print_state_change==P_Q_t_remaining)) << endl;
				}
			}
			catch(Exception& e)
			{
				m_bExitReq = true;
				cout << e << endl;
				throw e;
			}
		}
		if (m_model.TimeNext() == DBL_MAX)
		{
			cout << "All models are passive." << endl;
			cout << m_menu;
			m_bPauseReq = true; 
		}
		else if (m_StepMode == STEP_BY_ONE_SCHEDULE || m_model.TimeCurrent() >= m_ending_time) {  // stop looping
			m_bPauseReq = true; 
			cout << m_menu;
		}
	}
}
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void SRTEngine::Pause() 
{
	m_bPauseComp = false;
	m_bPauseReq = true;
	while(!m_bPauseComp)
		SLEEP(10);
}

void SRTEngine::KillLoop()
{
	m_bExitReq = true; // making it true 
	while(!m_bExitComp)
		SLEEP(10);
}

void SRTEngine::Loop()
{
	unsigned long PrevAniClock = milliseconds();
	while (!m_bExitReq)
	{
		if (!m_bPauseReq)
		{
			if (m_ExtX.size() > 0)// external input 
				TransmitExtEvent(); //-- (6)
			else {
				if(m_time_scale == DBL_MAX){ // pure virtual simulation, that is, discrete event simulation
					if(m_model.TimeNext() != DBL_MAX)
						Model::t_Current = m_model.TimeNext(); // jump to the tN
					ScheduleExecution(); 
				}
				else {
					unsigned long nowClock = milliseconds();
					double delta = (((double)(nowClock - m_PreviousClockTime)) / 1.0E+3) * m_time_scale;
					
					if(delta > 1E-05)  
					{
						Model::t_Current += delta;
						/*if(m_StepMode == STEP_BY_ONE_SCHEDULE && m_model.TimeNext() == DBL_MAX) 
							m_bPauseReq = false;
						else */ //  for passive model & waiting user input 
						double diff = m_model.TimeNext() - Model::t_Current;
						//if( diff < -1E-03) // too late
						//	cout << "VIOLATING real-time constraint of 1E-03 by " << diff << "!" << endl;
						if( diff < 1E-03) { // schedule time reached!
							Model::t_Current = m_model.TimeNext(); // adjust to tN
							ScheduleExecution();  
						}
						m_PreviousClockTime = nowClock; // update previous clock time
					} 
					{
						double ani_delta = (((double)(nowClock - PrevAniClock)) / 1.0E+3);
						if(!m_bPauseReq && b_acs && ani_delta > m_animation_interval){
							AnimationExecution();
							PrevAniClock = nowClock;
						}
					}
				}
			}
			//if(m_bigmouth) std::cout << "Time : " << Model::t_Current << "\n";
		}
		if(m_bPauseReq == true)
		{
			m_bPauseComp = true;
			SLEEP(100); // slow checking
		}
		else 
		{
			if(m_time_scale == DBL_MAX) // fastest run
				SLEEP(0);
			else // not fastest run
				SLEEP(10);
		}
	}	
//-- simulation exception 
	m_bExitComp = true;
}

void SRTEngine::PrintSettings() const
{
	cout << "scale factor: ";
	if( GetTimeScale()==DBL_MAX) cout << "inf" << endl;
	else cout << GetTimeScale() << endl;

	cout << ((m_StepMode==STEP_BY_ONE_SCHEDULE) ? "step-by-step mode" 
												: "run-through mode") << endl;
	cout << "current time: " << Model::TimeCurrent() << endl;
	cout << "simulation ending time: " << GetEndingTime() << endl;
	cout << "discrete transition printing mode:";
	if( print_state_change==P_NONE) cout << "none" << endl;
	else if(print_state_change==P_Q_t_elapsed) cout << "(state, elapsed_time, schedule_time_span)" << endl;
	else cout << "(state, schedule_time_span, remaining_time)" << endl;

	cout << "animation mode: ";
	if(b_acs) cout << " on and interval= " << m_animation_interval << endl;
	else cout << "off" << endl;
}

void SRTEngine::RunConsoleMenu()
{
	m_menu="scale, run, step, [p]ause, pause_at, reset, rerun, [i]nject, dtmode, ctmode, \nprint, cls, [e]xit \n>";

	string input;
	while(1)
	{
		cout << m_menu;
		cin >> input;
		
		if(input=="scale")
		{	double value;
			cout << "current scale value: " << GetTimeScale() << " new value > "; 
			cin >> value; 
			 
			if( value-1.0E-06 < 0 || value > 1.0E+06)
				SetTimeScale(DBL_MAX); 
			else 
				SetTimeScale(value);
		}
		else if(input == "run") Run();
		else if(input == "step") Step();
		else if(input == "pause" || input == "p") Pause();
		else if(input == "pause_at"){ Time at_time;
			cout << "ending time > "; cin >> at_time;  
			if( Model::TimeCurrent() <= at_time )
				SetEndingTime(at_time); 
			else 
				cout << "Ending time should be greater than or equal to " <<  Model::TimeCurrent() << endl;
		}
		else if(input=="reset") Reset();
		else if(input=="rerun") Rerun();
		else if (input == "inject" || input =="i" ){
			if(call_back_method==NULL)
				cout << "The call back function is not defined!" << endl;
			else {
				PortValue pv = call_back_method();
				Inject(pv); // PortValue(InputPort("in"), new Client(60)));
			}
		}
		else if (input == "dtmode"){
			cout << "options: none tr te >";
			cin >> input;  
			if(input == "none")  
				Set_dtmode(SRTEngine::P_NONE);
			else if(input == "tr")
				Set_dtmode(SRTEngine::P_Q_t_remaining);
			else if(input == "te")
				Set_dtmode(SRTEngine::P_Q_t_elapsed);
			else
				cout << "Invalid input!" << endl;

		}else if(input == "ctmode")
		{
			cout << "options: none avi >";
			cin >> input;  
			if (input == "none")
				SetAnimationFlag(false); // for animation;
			else { double ai=::atof(input.data());
				if(ai-1E-2 > 0) {
					SetAnimationFlag(true);
					SetAnimationInterval(ai);
				}else
					cout << "Invalid input!" << endl;
			}
		}
		else if(input == "print") {
			cout << "options: q cpl s > ";
			cin >> input;  
			if(input=="q")
				cout << m_model.Get_q(print_state_change==P_Q_t_remaining) << " at " << Model::TimeCurrent() << endl;
			else if(input == "cpl") {
				if(m_model.IsKindOf(ODEVS_RT(Coupled)))
					((Coupled&)m_model).PrintCouplings();
				else
					cout << "Model is not coupled DEVS!" << endl;
			}else if(input == "s")
				PrintSettings();
			else
				cout << "Invalid input!" << endl;
		}else if(input == "cls")
			CLEAR_SCREEN();
		else if (input =="e" || input == "exit")
			break;
	}
	KillLoop();
}